#!/usr/bin/env python3
from pwn import *
import struct
import sys
from collections import namedtuple

# --- Configuration ---
file_path = './labyrinth'
# Assuming 0x3036 is relative to the base, so absolute is 0x100000 + 0x3036
absolute_target_jmp_dest = 0x3036
assumed_base_addr = 0x100000
section_name = '.text'
offset_add_value = 0x13b6
ignored_offset_add_value = 3
direction_offset_base = 0x13b6 # Value to add to address_to_update for direction calc
direction_offset_sub = 6      # Value to subtract for direction calc
target_adds_count = 4 # Keep only the last 4 non-ignored ADD R15, imm32
direction_map = [0, 1, 2, 3] # 0: L, 1: D, 2: U, 3: R

# Instruction patterns (Opcode [+ ModRM] [+ Immediate Placeholder length])
PATTERNS = {
    # Name: (prefix_bytes, immediate_length)
    "JMP_REL32":       (b'\xE9', 4),
    "ADD_R15_IMM32":   (b'\x49\x81\xC7', 4),
    "ADD_R15_IMM8":    (b'\x49\x83\xC7', 1),
    "MOV_R14D_IMM32":  (b'\x41\xBE', 4),
    "MOV_MEMR15_R14D": (b'\x45\x89\x37', 0), # No immediate
    # "MOV_R15_R13": (b'\x4D\x89\xEF', 0), # Removed as per request
}

# Store info about each found instruction
InstructionInfo = namedtuple("InstructionInfo", ["addr", "pattern_name", "immediate", "length"])
# Store results for ignored blocks
IgnoredUpdateInfo = namedtuple("IgnoredUpdateInfo", ["addr_update", "val_update", "direction"])


# --- Helper Functions ---

def find_instructions(data, start_addr, patterns):
    """Finds all occurrences of specified instruction patterns in data."""
    found = []
    data_len = len(data)
    prefix_map = {pattern[0]: (name, pattern[1]) for name, pattern in patterns.items()}
    prefixes = list(prefix_map.keys())

    i = 0
    while i < data_len:
        matched = False
        for prefix in prefixes:
            prefix_len = len(prefix)
            # Check if prefix matches at current position
            if i + prefix_len <= data_len and data[i : i + prefix_len] == prefix:
                name, imm_len = prefix_map[prefix]
                instr_len = prefix_len + imm_len
                current_addr = start_addr + i

                # Check if the full instruction fits in the data
                if i + instr_len <= data_len:
                    immediate = None
                    if imm_len > 0:
                        imm_bytes = data[i + prefix_len : i + instr_len]
                        try:
                            if imm_len == 1:
                                immediate = struct.unpack('<b', imm_bytes)[0] # Signed byte
                            elif imm_len == 4:
                                immediate = struct.unpack('<i', imm_bytes)[0] # Signed dword
                            else:
                                log.warning(f"Unsupported immediate length {imm_len} for {name} at {hex(current_addr)}")
                        except struct.error as e:
                            log.warning(f"Struct unpack error for {name} at {hex(current_addr)} (len {imm_len}): {e}")

                    found.append(InstructionInfo(addr=current_addr, pattern_name=name, immediate=immediate, length=instr_len))
                    i += instr_len # Advance past the matched instruction
                    matched = True
                    break # Move to next position after match
                else:
                    # Prefix matches but instruction is cut off
                    log.debug(f"Pattern {name} prefix match at {hex(current_addr)} but cut off.")
                    # Don't advance fully, just move to next byte to allow overlap if needed
                    i += 1
                    matched = True
                    break # Exit inner loop, continue outer loop

        if not matched:
            i += 1 # Advance if no pattern prefix matched at this byte

    return found


def find_delimiters(instructions, target_jmp_addr):
    """Filters found instructions to find JMP delimiters."""
    delimiters = []
    log.info(f"Searching for JMP {hex(target_jmp_addr)} delimiters among instructions...")
    for instr in instructions:
        if instr.pattern_name == "JMP_REL32" and instr.immediate is not None:
            # Calculate target address: addr + length + relative_offset
            address_after_jmp = instr.addr + instr.length
            potential_target = address_after_jmp + instr.immediate
            if potential_target == target_jmp_addr:
                log.debug(f"Found delimiter JMP at {hex(instr.addr)} targeting {hex(potential_target)}")
                delimiters.append(instr.addr)
    delimiters.sort()
    if delimiters:
         log.success(f"Found {len(delimiters)} delimiter JMPs targeting {hex(target_jmp_addr)}.")
    else:
        log.warning(f"No JMP instructions targeting {hex(target_jmp_addr)} found.")
    return delimiters


def process_section(section_instructions, section_start_addr, section_end_addr, full_text_data, text_vaddr):
    """Processes instructions within a section to find target offsets and ignored block updates."""
    log.info(f"Processing section {hex(section_start_addr)} - {hex(section_end_addr)}")

    gate_updates = [] # List of IgnoredUpdateInfo tuples
    target_offsets = [] # List of modified offsets for the last N ADD R15, imm32

    # Filter instructions for the current section and map by address
    instr_map = {instr.addr: instr for instr in section_instructions if section_start_addr <= instr.addr < section_end_addr}
    # Keep track of ADD R15, imm32 addresses that are part of ignored blocks
    ignored_add_imm32_addrs = set()

    # --- Pass 1: Identify Ignored Blocks ---
    log.debug("Searching for 'updates' blocks...")
    # Get potential starting instructions (ADD R15, imm32) sorted by address
    potential_starts = sorted([instr for instr in instr_map.values() if instr.pattern_name == "ADD_R15_IMM32"], key=lambda x: x.addr)

    for start_instr in potential_starts:
        addr = start_instr.addr
        if addr in ignored_add_imm32_addrs: # Already processed
            continue

        # Calculate expected addresses of subsequent instructions in the *modified* pattern
        # Pattern: ADD imm32 -> ADD imm8 -> MOV R14D, imm32 -> MOV [R15], R14D
        addr_add_imm8 = addr + start_instr.length
        instr_add_imm8 = instr_map.get(addr_add_imm8)

        if not (instr_add_imm8 and instr_add_imm8.pattern_name == "ADD_R15_IMM8"):
            continue # Doesn't match start of pattern

        addr_mov_r14d = addr_add_imm8 + instr_add_imm8.length
        instr_mov_r14d = instr_map.get(addr_mov_r14d)

        if not (instr_mov_r14d and instr_mov_r14d.pattern_name == "MOV_R14D_IMM32"):
             continue

        addr_mov_mem  = addr_mov_r14d + instr_mov_r14d.length
        instr_mov_mem = instr_map.get(addr_mov_mem)

        # Check if the final parts of the pattern match
        if (instr_mov_mem and instr_mov_mem.pattern_name == "MOV_MEMR15_R14D"):
            # --- Found Ignored Block ---
            log.debug(f"  Found ignored block starting at {hex(addr)}")

            original_offset = start_instr.immediate
            value_to_update = instr_mov_r14d.immediate

            if original_offset is not None and value_to_update is not None:
                address_to_update = original_offset + ignored_offset_add_value

                # --- Calculate and Resolve Direction ---
                direction_char = -1 # Default/error value
                direction_addr = address_to_update + direction_offset_base - direction_offset_sub
                log.debug(f"    Calculated direction address: {hex(direction_addr)}")

                # Convert virtual address to offset in text_data
                direction_offset_in_data = direction_addr - text_vaddr
                if 0 <= direction_offset_in_data < len(full_text_data):
                    direction_byte = full_text_data[direction_offset_in_data]
                    if 0 <= direction_byte < len(direction_map):
                        direction_char = direction_map[direction_byte]
                        log.debug(f"    Read direction byte {hex(direction_byte)} -> '{direction_char}'")
                    else:
                        log.warning(f"    Invalid direction byte {hex(direction_byte)} found at {hex(direction_addr)}")
                else:
                    log.warning(f"    Direction address {hex(direction_addr)} is outside .text bounds [{hex(text_vaddr)}, {hex(text_vaddr+len(full_text_data))})")
                # --- End Direction Calculation ---

                gate_updates.append(IgnoredUpdateInfo(
                    addr_update=address_to_update + 0x13b6,
                    val_update=value_to_update + 0x13b6,
                    direction=direction_char
                ))
                ignored_add_imm32_addrs.add(addr) # Mark this ADD as processed
                log.debug(f"    -> Update Addr: {hex(address_to_update)}, Value: {hex(value_to_update)}, Direction: {direction_char}")

            else:
                 log.warning(f"  Could not extract immediate values for ignored block at {hex(addr)}")
        # else: Pattern did not fully match starting here

    # --- Pass 2: Identify Target ADD R15, imm32 (last N non-ignored) ---
    log.debug(f"Identifying last {target_adds_count} target 'ADD R15, imm32' instructions...")
    potential_targets = []
    for instr in potential_starts: # Iterate again through all ADD R15, imm32 in section
        if instr.addr not in ignored_add_imm32_addrs:
            if instr.immediate is not None:
                potential_targets.append(instr)
            else:
                 log.warning(f"  Non-ignored ADD R15, imm32 at {hex(instr.addr)} missing immediate? Skipping.")

    # Get the last N targets based on address (list is already sorted)
    actual_targets = potential_targets[-target_adds_count:]

    for target_instr in actual_targets:
        modified_offset = target_instr.immediate + offset_add_value
        target_offsets.append(modified_offset)
        log.debug(f"  Found target ADD R15, imm32 at {hex(target_instr.addr)}, original offset: {hex(target_instr.immediate)}, modified: {hex(modified_offset)}")

    log.info(f"  Finished processing section. Found {len(gate_updates)} ignored updates and {len(target_offsets)} target offsets.")
    return gate_updates, target_offsets


# --- Main Execution ---
try:
    # Load ELF and set context/base address
    elf = ELF(file_path, checksec=False)
    context.binary = elf
    elf.address = assumed_base_addr
    log.info(f"ELF loaded. Assuming base address: {hex(elf.address)}")

    # Get .text section
    text_section = elf.get_section_by_name(section_name)
    if not text_section:
        log.error(f"Section '{section_name}' not found in {file_path}")
        sys.exit(1)

    text_vaddr = text_section['sh_addr'] # Base address relative to elf.address
    text_data = text_section.data()
    text_size = len(text_data)
    text_end_vaddr = text_vaddr + text_size
    log.info(f"Found '{section_name}' section @ {hex(text_vaddr)}, size {hex(text_size)}")

    # 1. Find *all* relevant instructions in the .text section once
    log.info("Scanning .text section for relevant instruction patterns...")
    all_instructions = find_instructions(text_data, text_vaddr, PATTERNS)
    log.success(f"Found {len(all_instructions)} relevant instructions in total.")

    # 2. Find the delimiter addresses using the ABSOLUTE target jump address
    delimiter_addrs = find_delimiters(all_instructions, absolute_target_jmp_dest)

    if not delimiter_addrs:
        sys.exit(0)

    # 3. Process each section defined by the delimiters
    print(f"\n--- Processing Sections (Delimiter: JMP {hex(absolute_target_jmp_dest)}) ---")
    num_delimiters = len(delimiter_addrs)
    all_section_results = [] # Store results for each section

    for i in range(num_delimiters):
        section_start_addr = delimiter_addrs[i]
        # Section ends right before the next delimiter, or at the end of .text
        if i + 1 < num_delimiters:
            section_end_addr = delimiter_addrs[i+1]
        else:
            section_end_addr = text_end_vaddr # Last section goes to the end

        print(f"\n--- Section {i+1} ---")
        print(f"  Range: [{hex(section_start_addr + 5)}, {hex(section_end_addr)})")

        # Process this section (pass necessary data for direction lookup)
        gate_updates, target_offsets = process_section(
            all_instructions, # Pass all instructions for filtering
            section_start_addr,
            section_end_addr,
            text_data,      # Pass full .text data
            text_vaddr      # Pass .text base virtual address
        )
        all_section_results.append({
            "section": i + 1,
            "start": section_start_addr + 5,
            "end": section_end_addr,
            'updates': gate_updates,
            "target": target_offsets
        })

        if gate_updates:
            print("  Ignored Block Updates (Address, Value, Direction):")
            for update in gate_updates:
                print(f"    ({hex(update.addr_update)}, {hex(update.val_update)}, '{update.direction}')")
        else:
            print("  No ignored block updates found.")

        if target_offsets:
            print(f"  Target ADD R15 Offsets (+ {hex(offset_add_value)}):")
            print(f"    {[hex(o) for o in target_offsets]}")
        else:
            print(f"  No target ADD R15 offsets found (expected last {target_adds_count}).")

    
    adj_list = {} # node : [neigbours]
    starts = [0x13c5] # node start addr

    for i in range(len(all_section_results)):
        s = all_section_results[i]
        starts.append(s['start'])
    
    starts.append(0x3036)
    adj_list[0] = [16, 0, 0, 0]
    adj_list[39] = []
    
    for i in range(len(all_section_results)):
        s = all_section_results[i]
        adj_list[i+1] = []
        for t in s['target']:
            adj_list[i+1].append(starts.index(t))
    
    adj_list[16].append(0)
    print(adj_list)
    gate = {}
    gates = {}

    for i in range(len(all_section_results)):
        s = all_section_results[i]
        if s['updates']:
            gate[i+1] = []
            gates[i+1] = (2 ** len(gates))
            for j in range(len(s['updates'])//2):
                k = 0
                ignore = s['updates'][j]
                while k < len(starts):
                    if ignore.addr_update < starts[k]:
                        gate[i+1].append((k-1, ignore.direction, starts.index(ignore.val_update)))
                        # alt_adj_list[k-1][ignore.direction] = starts.index(ignore.val_update)
                        print(f"Changed {k-1} direction {ignore.direction} to point to {starts.index(ignore.val_update)} ({hex(ignore.val_update)})")
                        break
                    k += 1
    
    # target is 20

    def neighbours_by_state(node, state):
        neighbours = adj_list[node].copy()
        for s in state:
            for change in gate[s]:
                if change[0] == node:
                    neighbours[change[1]] = change[2]
        if state != 0: print(node, state, neighbours)
        return neighbours


    visited = set()
    q = [("", 0, set())]
    dirs = ["L", "D", "U", "R"]
    while len(q) > 0:
        x = q[0]
        q = q[1:]

        s = x[2].copy()
        bit = 0
        for g in s:
            bit += gates[g]
        visited.add((x[1], bit))
        print(f"Added {x[1]} {bin(bit)}")

        if x[1] in gate.keys():
            if x[1] in s:
                s.remove(x[1])
            else:
                s.add(x[1])
        bit = 0
        for g in s:
            bit += gates[g]

        if x[1] != 38:
            print(x[0], x[1])
        if x[1] == 19:
            print("-"*30)
            print("SOLUTION: " + x[0] + "D")
            break

        
        
        neighbours = neighbours_by_state(x[1], s)
        print(neighbours)
        for i in range(len(neighbours)):
            n = neighbours[i]
            if (n, bit) not in visited:
                q.append((x[0]+dirs[i], n, s))
                print(f"Pushed {(x[0]+dirs[i], n)}")
        


except FileNotFoundError:
    log.error(f"Error: File not found at {file_path}")
    sys.exit(1)
except Exception as e:
    log.error(f"An unexpected error occurred: {e}")
    import traceback
    traceback.print_exc()
    sys.exit(1)